;Ein Vektor Type, wird benutzt fr Punkte!
; hat die Felder x und y, welche Floats sind
; in x und y werden die Koordinaten des Punktes gespeichert
Type Vector2
	Field x#
	Field y#
End Type

;Diese Funktion erstellt einen neuen Vektor!
; Parameter: x und y
; die Parameter werden in der Typeinstanz gespeichert
; der Vektor wird zurckgeliefert
Function newVector2.Vector2(x#, y#)
	Local v.Vector2=New Vector2
	
	v\x=x
	v\y=y
	
	Return v
End Function

;Maximale Zahl fr Punkte in der Fraktalen Map
Const MAX_FRACTAL_POINTS=128

;Fraktale Map,
; hat eine Liste von Punkten (Vector2)
; hat einen Zhler, wieviel Punkte schon in der Map sind
Type TFractalMap
	Field p.Vector2[MAX_FRACTAL_POINTS]
	Field count
End Type

;Fgt einen Punkt zur Fraktalmap hinzu,
; Parameter m, p
; m ist die Map in der der Punkt p eingefgt werden soll
; bricht ab, falls schon zuviele Punkte in der Map sind
Function addPointToMap(m.TFractalMap, p.Vector2)
	If m\count>=MAX_FRACTAL_POINTS
		Return
	EndIf
	
	m\p[m\count]=p
	m\count=m\count+1
End Function

;Fgt einen Punkt zur Fraktalmap an der Position i hinzu
; Parameter: m, p, i
; Fgt den Punkt p an der Stelle i in die Map m ein
; bricht ab, falls schon zuviele Punkte in der Map sind
; verschiebt alle anderen Punkte um eins nach vorne
Function insertPointToMap(m.TFractalMap, p.Vector2, i)
	If m\count>=MAX_FRACTAL_POINTS
		Return
	EndIf
	
	For j=m\count To i+1 Step -1
		m\p[j]=m\p[j-1]
	Next
	
	m\p[i]=p
	
	m\count=m\count+1
End Function

;Zeichnet die Fraktale Map
; Parameter: m, fill
; zeichnet die Map m gefllt oder leer, je nach Parameter
Function drawFractalMap(m.TFractalMap, fill=False)
	LockBuffer BackBuffer()
	
	If Not fill
		For i=1 To m\count-1
			Line m\p[i]\x, m\p[i]\y, m\p[i-1]\x, m\p[i-1]\y
		Next
	Else
		For i=1 To m\count-1
			fillMapSegment(m\p[i-1]\x, m\p[i-1]\y, m\p[i]\x, m\p[i]\y)
		Next
	EndIf
	
	UnlockBuffer BackBuffer()
End Function

;Fllt ein MapSegment
; Parameter x0, y0, x1, y1
; die vier Parameter sind (in Zweierpaaren) die Koordinaten
; der Linienpunkte, die nach unten hin gefllt werden soll
Function fillMapSegment(x0, y0, x1, y1)
	Local steep
	Local tmp, dX, dY, err, yStep, y
	
	If Abs(y1-y0)>Abs(x1-x0)
		steep=True
		
		tmp=x0
		x0=y0
		y0=tmp
		
		tmp=x1
		x1=y1
		y1=tmp
	Else
		steep=False
	EndIf
	
	If x0>x1
		tmp=x0
		x0=x1
		x1=tmp
		
		tmp=y0
		y0=y1
		y1=tmp
	EndIf
	
	dX=x1-x0
	dY=Abs(y1-y0)
	err=dX/2
	y=y0
	
	If y0<y1
		yStep=1
	Else
		yStep=-1
	EndIf
	
	If steep
		For x=x0 To x1
			Line y, x, y, GraphicsHeight()-1
			
			err=err-dY
			
			If err<0
				y=y+yStep
				
				err=err+dX
			EndIf
		Next
	Else
		For x=x0 To x1
			Line x, y, x, GraphicsHeight()-1
			
			err=err-dY
			
			If err<0
				y=y+yStep
				
				err=err+dX
			EndIf
		Next
	EndIf
End Function

;Iteriert die Map
; Parameter: m, depth, maxStep
; Iteriert Map m (= bringt sie voran), Iterationstiefe ist
; depth, maximale Verschiebung der einzelnen Schritte ist
; maxStep
Function iterateMap(m.TFractalMap, depth, maxStep#=150)
	Local mx#, my#, stp#
	
	For d=0 To depth-1
		For i=0 To m\count-2
			mx=((m\p[i+1]\x-m\p[i]\x)/2.)+m\p[i]\x
			my=((m\p[i+1]\y-m\p[i]\y)/2.)+m\p[i]\y
			
			stp=maxStep*(1./(d+1))^2
			
			my=my+Rnd(0,stp*2)-stp
			
			insertPointToMap(m, newVector2(mx, my), i+1)
			i=i+1
		Next
	Next
End Function

;/*
; *
; */

;Schreibt eine Map in ein Bild
; Parameter: m, fill
; Zeichnet die Map m in ein neu erstelltes Bild (nach
; Bildschirmgrsse) und fllt sie je nach Parameter
; liefert das Bild zurck
Function writeMapToImage(m.TFractalMap, fill=True)
	Local r=ColorRed()
	Local g=ColorGreen()
	Local b=ColorBlue()
	
	img=CreateImage(GraphicsWidth(), GraphicsHeight())
	MaskImage img,0,0,0
	
	SetBuffer ImageBuffer(img)
	
	Color 0,0,0
	Rect 0, 0, GraphicsWidth(), GraphicsHeight()
	
	Color r,g,b
	drawFractalMap(m, fill)
	
	SetBuffer BackBuffer()
	
	Return img
End Function

;Prft die Map auf Collision an einem Punkt
; Parameter: img, x, y
; Prft ob die Map (gespeichert im Bild img) an der Stelle x/y
; gesetzt oder leer ist
; liefert false zurck, falls x/y nicht im Bild sind
Function checkMapCollision(img, x, y)
	If Not RectsOverlap(x,y,1,1, 0,0, GraphicsWidth(), GraphicsHeight())
		Return False
	EndIf
	
	Local col=ReadPixelFast(x, y, ImageBuffer(img))
	
	r=(col And $ff0000)/$10000
	G=(col And $ff00)/$100
	b=(col And $ff)
	
	If r=0 And G=0 And b=0
		Return False
	EndIf
	
	Return True
End Function

;Explodiert die Karte
; Parameter img, x, y, r
; Explodiert die Karte (gespeichert im Bild img) an der Stelle
; x/y, die Explosion hat den Radius r
Function explodeMap(img, x, y, r)
	SetBuffer ImageBuffer(img)
	
	Color 0,0,0
	Oval x-r, y-r, 2*r, 2*r
	
	SetBuffer BackBuffer()
End Function

;/*
; *
; */

;Ein Projektil Type
; Hat die Felder x/y und vx/vy,
; x/y dient zur Speicherung der Position
; vx/vy dient zur Speicherung der Geschwindigkeit
Type TProjectile
	Field x#
	Field y#
	
	Field vx#
	Field vy#
End Type

;Erstellt ein neues Projektil
; Parameter: x, y
; die Funktion erstellt ein neues Projektil an der Stelle
; x/y, die Anfangsgeschwindigkeit wird auf 0/0 gesetzt
Function newProjectile.TProjectile(x#, y#)
	Local p.TProjectile=New TProjectile
	
	p\x=x
	p\y=y
	
	p\vx=0
	p\vy=0
	
	Return p
End Function

;Schadensradius fr Projektile
Const dmgRad#=40

;Bearbeitet die Projektile
; Parameter: img
; Erneuert die Position/Geschwindigkeit aller Projektile
; berprft und berechnet Einschlge in der Map (gespeichert
; im Bild img)
; lscht berflssige Projektile
Function updateProjectiles(img)
	Local p.TProjectile
	
	For p=Each TProjectile
		If p\x<0 Or p\y<0 Or p\x>=GraphicsWidth() Or p\y>=GraphicsHeight()
			calculateDamage(p)
			
			Delete p
		ElseIf checkMapCollision(img, p\x, p\y)
			explodeMap(img, p\x, p\y, dmgRad/1.5)
			calculateDamage(p)
			
			Delete p
		Else
			p\vy=p\vy+0.1
			
			p\x=p\x+p\vx
			p\y=p\y+p\vy
			
			Oval p\x-2,p\y-2,4,4
			;Oval p\x-dmgRad, p\y-dmgRad, 2*dmgRad, 2*dmgRad, 0
		EndIf
	Next
End Function

;Berechnet den Schaden eines Projektils
; Parameter: p
; Berechnet den Schaden fr beide Spieler, anhand des
; Projektileinschlags!
Function calculateDamage(p.TProjectile)
	Local maxDmg#=50
	Local minDmg#=5
	
	Local m#=(minDmg-maxDmg)/dmgRad
	
	Local dist#, dx#, dy#
	
	dx=p\x-p1X
	dy=p\y-p1Y
	
	dist=Sqr(dx*dx+dy*dy)
	
	If dist<dmgRad
		health1=health1-m*dist-maxDmg
	EndIf
	
	dx=p\x-p2X
	dy=p\y-p2Y
	
	dist=Sqr(dx*dx+dy*dy)
	
	If dist<dmgRad
		health2=health2-m*dist-maxDmg
	EndIf
End Function

;/*
; *
; */

;SpielerPositionen
Global p1X#, p1Y#
Global p2X#, p2Y#

;Spielerwinkel
Global a1#, a2#

;Spielkarte
Global map.TFractalMap
Global image

;Spielzug
Global turn

;Spielerleben
Global health1#, health2#

;Verarbeitungsvariabeln
Local p.TProjectile
Local l#, dx#, dy#, vx#, vy#, t#, ey#

;Menuvariabeln
Local humanPlayer=False
Global menuActive=True
Local menuCase=0

;Setzt das Spiel zurck
; lscht alle Type Instanzen
; generiert eine neue Map
; schreibt diese in das MapBild
; setzt die Spieler zufllig aufs Feld
; startet das Spiel
Function reset()
	If image<>0
		FreeImage(image)
		image=0
	EndIf
	
	Delete Each TFractalMap
	Delete Each Vector2
	Delete Each TProjectile
	
	map=New TFractalMap
	
	addPointToMap(map, newVector2(0, 400))
	addPointToMap(map, newVector2(800, 400))
	
	SeedRnd(MilliSecs())
	iterateMap(map, 10, 200)
	
	Color 0,200,0
	image=writeMapToImage(map)
	
	p1X=Rand(50, 350)
	p2X=Rand(450, 750)
	
	a1=90
	a2=-90
	
	health1=100
	health2=100
	
	turn=0
End Function

;Gibt die Y Koordinate einer X Koordinate zurck
; Parameter: img, x
; findet die erste gefllte Y Koordinate in der Map die
; auf der gesuchten X-Position liegt
Function getYonMap(img, x#)
	For y=0 To 600
		If checkMapCollision(img, x, y)
			Return y
		EndIf
	Next
	
	Return 600
End Function

;Startet das Menu
; lscht alle Typeinstanzen
; aktiviert das Menu
Function startMenu()
	If image<>0
		FreeImage(image)
		image=0
	EndIf
	
	Delete Each TFractalMap
	Delete Each Vector2
	Delete Each TProjectile
	
	menuActive=True
End Function

;/*
; *
; */

;Grafik Kontext
Graphics 800,600,0,2
SetBuffer BackBuffer()

;Bilder werden geladen
; danke Bruzard fr das Tank Bild,
; ich hoffe er nimmt es mir nicht bel, dass ich es ver-
; unstaltet und zerschnippselt habe :/
Local tank1=LoadImage("Tank_Bruz.png");LoadImage("Tank.png")
MaskImage tank1, 255, 255, 255
MidHandle tank1

;Spiegelt das Bild fr den zweiten Spieler,
; ja, ich bin faul.
Local tank2=CopyImage(tank1)
ScaleImage tank2, -1, 1
MidHandle tank2

;Timer, zur Verhinderung von 100% Systemlast
Local timer=CreateTimer(60)

;Hauptschleife
While Not KeyHit(1)
	;Abfrage, Spiel oder Menu
	If Not menuActive
		;Zeichnet die Map
		DrawImage image, 0,0
		
		;Setzt die Spieler auf die map
		p1Y=getYonMap(image, p1X)
		p2Y=getYonMap(image, p2X)
		
		;berprft, wer am Zug ist
		Select turn
			;Spieler 1 am Zug
			Case 0
				;Zeichnet die Ziellinie fr Spieler1
				Color 125, 125, 125
				Line p1X+10, p1Y-17, MouseX(), MouseY()
				
				;Abschuss
				If MouseHit(1)
					p=New TProjectile
					
					p\x=p1X+10+Sin(a1)*20
					p\y=p1Y-17-Cos(a1)*20
					
					dx=MouseX()-p1X-10
					dy=MouseY()-p1Y+17
					
					l=Sqr(dx*dx+dy*dy)/50
					
					p\vx=Sin(a1)*l
					p\vy=-Cos(a1)*l
					
					turn=1-turn
				EndIf
				
				;Winkelberechnung anhand der Maus
				a1=ATan2(MouseY()-p1Y+17, MouseX()-p1X-10)+90
				
				;Winkeleinschrnkung
				If a1<10
					a1=10
				EndIf
				If a1>90
					a1=90
				EndIf
				
			;Spieler 2 am Zug
			Case 1
				;Zeichnet die Ziellinie fr Spieler 2
				Color 125, 125, 125
				Line p2X-10, p2Y-17, MouseX(), MouseY()
				
				;Gegen Mensch oder Computer?
				If humanPlayer
					;Abschuss
					If MouseHit(1)
						p=New TProjectile
						
						p\x=p2X-10+Sin(a2)*20
						p\y=p2Y-17-Cos(a2)*20
						
						dx=MouseX()-p2X+10
						dy=MouseY()-p2Y+17
						
						l=Sqr(dx*dx+dy*dy)/50
						
						p\vx=Sin(a2)*l
						p\vy=-Cos(a2)*l
						
						turn=1-turn
					EndIf
					
					;Winkelberechnung anhand von Maus
					a2=ATan2(MouseY()-p2Y+17, MouseX()-p2X+10)+90
					
					;Winkeleinschrnkung
					If a2<-90 Or a2>180
						a2=-90
					EndIf
					If a2>10
						a2=10
					EndIf
				Else
					;Schussvektor des Computers
					dx=p1X-p2X
					dy=-dx
					
					l=Sqr(dx*dx+dy*dy)/50
					vx=Sin(a2)*l
					vy=Cos(a2)*l
					
					;Berechnung der Schusszeit
					t=(p1X-p2X)/vx
					
					;Berechnung des Einschlagspunkts
					ey=p2Y-vy*t+0.05*t^2
					
					;Einstellung des Schusswinkels
					If ey>(p1Y+20)
						a2=a2+1
					Else
						a2=a2-1
					EndIf
					
					;Winkeleinschrnkung
					blocked=False
					If a2<-90 Or a2>180
						a2=-90
						blocked=True
					EndIf
					If a2>10
						a2=10
						blocked=True
					EndIf
					
					;Abschuss
					If Abs(ey-p1Y-20)<10 Or blocked
						p=New TProjectile
						
						p\x=p2X-10+Sin(a2)*20
						p\y=p2Y-17-Cos(a2)*20
						
						p\vx=Sin(a2)*l
						p\vy=-Cos(a2)*l
						
						turn=1-turn
					EndIf
				EndIf
		End Select
		
		;Spieleroutine, erneuert die Projektile
		updateProjectiles(image)
		
		;Zeichnet das Spielgeschehen
		Color 255,255,255
		DrawImage tank1, p1X, p1Y-10
		DrawImage tank2, p2X, p2Y-10
		
		Line p1X+10, p1Y-17, p1X+10+Sin(a1)*20, p1Y-17-Cos(a1)*20
		Line p2X-10, p2Y-17, p2X-10+Sin(a2)*20, p2Y-17-Cos(a2)*20
		
		Rect p1X-10, p1Y-40, 20, 6, 0
		Rect p2X-10, p2Y-40, 20, 6, 0
		
		Color 255, 0, 0
		Rect p1X-10, p1Y-40, (health1/100)*20, 6
		Rect p2X-10, p2Y-40, (health2/100)*20, 6
		
		;Tot eines der Spieler, zurck zum Menu
		If health1<=0 Or health2<=0
			startMenu()
			menuCase=0
		EndIf
	Else
		;Menuschleife
		Color 255, 255, 255
		
		;Menupunkte
		Text 368, 200, "Vs Human"
		Text 356, 300, "Vs Computer"
		Text 384, 400, "Exit"
		
		;Menuauswahl
		Rect 300, 185+menuCase*100, 200, 40,0
		
		;Mausauswahl
		If MouseX()>300 And MouseX()<500
			If MouseY()>180 And MouseY()<420
				menuCase=(MouseY()-150)/100
			EndIf
		EndIf
		
		;Tastaturauswahl
		menuCase=menuCase-KeyHit(200)+KeyHit(208)
		If menuCase<0
			menuCase=2
		EndIf
		If menuCase>2
			menuCase=0
		EndIf
		
		;Machwas
		If MouseHit(1) Or KeyHit(57) Or KeyHit(28)
			If menuCase=0
				reset()
				humanPlayer=True
			ElseIf menuCase=1
				reset()
				humanPlayer=False
			Else
				End
			EndIf
			
			menuActive=False
		EndIf
	EndIf
	
	;FPS Counter
	fpsCount=fpsCount+1
	If MilliSecs()-fpsTime>999
		fpsCur=fpsCount
		fpsCount=0
		fpsTime=MilliSecs()
	EndIf
	Text 10,10,fpsCur
	
	;Update Grafikkontext
	Flip 0
	WaitTimer(timer)
	Cls
Wend

;Beende das Programm
End
;~IDEal Editor Parameters:
;~C#Blitz3D